/**
  Digital Elevation Model (DEM) demo application.
  
  Reads a USGS "BIL" (raw 16-bit binary) elevation image, and 
  draws it onscreen as a vertex buffer object.
  
  Orion Sky Lawlor, olawlor@acm.org, 2008-04-15 (Public Domain)
*/
#include <GL/glew.h> /*<- for gl...ARB extentions.  Must call glewInit after glutCreateWindow! */
#include <iostream>
#include <GL/glut.h>
#include <math.h>
#include "osl/vec4.h" /*<- handy little vector class */
#include "osl/mat4.h"
using osl::mat4;
/* Convert a struct with .x, .y, .z fields into three arguments */
#define VEC3_TO_XYZ(v) (v).x,(v).y,(v).z


#include "ogl/image.h" /* (originally from) Nigel Stewart's GLT image library */
#include "ogl/image.cpp"

#include "ogl/glsl.h" /* Orion's tiny GLSL library */
#include "ogl/glsl.cpp"

#include "GL/glui.h" /* OpenGL User Interface library (for making a GUI) */

#ifndef M_PI
# define M_PI 3.14159265358979
#endif


float walking_speed=3.0;
int move_camera=1;
mat4 light_direction=mat4(1.0);


void oglExit(const char *where,const char *why) {
	printf("Fatal error> %s\n",why);
	exit(1);
}

/*********************** App Code **********************/
class usgs_bil_terrain {
public:
	usgs_bil_terrain(std::string basename);
	
	void draw(void);
	
	/* Return OpenGL 3D location of this terrain image pixel */
	vec3 get_xyz(int x,int y);
	int w,h;/* Pixel dimensions of our terrain image */
public:
	std::vector<int> heights; /* meters above sea level */
	
	/* Longitude / latitude / elevation */
	vec3 origin, x_axis, y_axis, z_axis;
};

usgs_bil_terrain::usgs_bil_terrain(std::string basename)
	:w(0), h(0)
{
	const char *where="usgs_bil_terrain::usgs_bil_terrain";
	FILE *f;

/* Read width and height from .hdr file */
	f=fopen((basename+".hdr").c_str(),"r");
	if (!f) {oglExit(where,"Can't open terrain file .hdr!\n");}
	char line[1000],kw[1000];
	while (NULL!=fgets(line,1000,f)) {
		int value;
		if (2==sscanf(line,"%s %d",kw,&value)) {
			if (kw==std::string("NROWS")) h=value;
			if (kw==std::string("NCOLS")) w=value;
		}
	}
	fclose(f);
	if (w==0 || h==0) {oglExit(where,"Can't parse terrain file .hdr!\n");}

/* Read pixel geometric sizes from .blw file */
	f=fopen((basename+".blw").c_str(),"r");
	if (!f) {oglExit(where,"Can't open terrain file .blw!\n");}
	double v[6];
	for (int i=0;i<6;i++)
		if (1!=fscanf(f," %lf ",&v[i]))
			oglExit(where,"Can't parse terrain file .blw!\n");
	fclose(f);
	x_axis=vec3(v[0],v[1],0);
	y_axis=vec3(v[2],v[3],0);
	z_axis=vec3(0,0,1);
	origin=vec3(v[4],v[5],0);
	
/* Read elevation values from .bil file */
	f=fopen((basename+".bil").c_str(),"rb");
	if (!f) {oglExit(where,"Can't open terrain file .bil!\n");}
	for (int y=0;y<h;y++)
	for (int x=0;x<w;x++) {
		unsigned char v[2];
		if (1!=fread(v,2,1,f)) oglExit(where,"Can't read terrain file .bil!\n");
		heights.push_back(v[0]+(v[1]<<8));
	}
	fclose(f);
}

vec3 usgs_bil_terrain::get_xyz(int x,int y)
{
	float z=heights[y*w + x];
	vec3 lle=x*x_axis+y*y_axis+z*z_axis + origin;
	
	float EARTH_RADIUS=6360000; /* approximate earth radius */
	float r=EARTH_RADIUS + lle.z; /* ASSUMES earth is a sphere!!! */
	float lat=lle.y*(M_PI/180.0), lon=lle.x*(M_PI/180.0);
	vec3 xyz=r*vec3(   
		cos(lat)*cos(lon),
		cos(lat)*sin(lon),
		sin(lat)  /* ASSUMES geocentric lat, not geodetic! */
	);
	
	return vec3(x*26.0, // 2" lon == 40e6*cos(65/180*pi)/360*2/3600
		-y*60.0, // 2" lat == 40e6/360*2/3600 (y increases to south)
		z
	);
}

void usgs_bil_terrain::draw(void) {

	static GLhandleARB prog=makeProgramObject(
		"varying vec4 color;\n"
		"void main(void) {\n"
		"	vec4 v=vec4(gl_Vertex.x,gl_Vertex.y,gl_Vertex.z,1);\n"
		"	color=v;\n"
		"	gl_Position = gl_ModelViewProjectionMatrix * v;\n"
		"}\n"
	,
		"varying vec4 color;\n"
		"void main(void) { gl_FragColor = fract(color.z*0.01); }\n"
	);
	glUseProgramObjectARB(prog);
	
	static GLuint vb=0, vb_len=0;; 
	if (vb==0) { /* set up the vertex buffer object */
		glGenBuffersARB(1,&vb);  /* make a buffer */
		glBindBufferARB(GL_ARRAY_BUFFER_ARB,vb);
		
		// Our vertex data is organized into a bunch of triangle strips.
		std::vector<vec3> vtx;
		for (int y=0;y<h-1;y++) {
			vtx.push_back(get_xyz(0,y));
			for (int x=0;x<w;x++) {
				vtx.push_back(get_xyz(x,y));
				vtx.push_back(get_xyz(x,y+1));
			}
			vtx.push_back(get_xyz(w-1,y+1));
		}
		vb_len=vtx.size();
		
		 /* Copy our vtx array (on the CPU) into our new GPU buffer */
		glBufferDataARB(GL_ARRAY_BUFFER_ARB,sizeof(vec3)*vb_len,
			&vtx[0],GL_STATIC_DRAW_ARB);
		glEnableClientState(GL_VERTEX_ARRAY);
 		glVertexPointer(3,GL_FLOAT,sizeof(vtx[0]), (void *)0);
		glBindBufferARB(GL_ARRAY_BUFFER_ARB,0); /* back to plain OpenGL */
	}
	/* Draw all our (GPU) points.  
	   This is about 3x faster than looping over 
	   vtx and calling glVertex again and again.
	*/
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,vb);
	glDrawArrays(GL_TRIANGLE_STRIP,0,vb_len);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB,0);
	
	glUseProgramObjectARB(0);
}

/* Orthonormal coordinate frame */
class ortho_frame {
public:
	/* Unit vectors pointing along axes of our frame.
		X cross Y is Z. (right-handed coordinate system)
	*/
	vec3 x,y,z;
	ortho_frame() :x(1,0,0), y(0,1,0), z(0,0,1) {}
	
	/* Reorient this coordinate frame by this far 
	   along in the X and Y axes.  "dx" is the distance
	   along the z axis to push the x axis; "dy" 
	   the distance along the z axis to push the y axis.
	*/
	void nudge(double dx,double dy) {
		x+=dx*z;
		y+=dy*z;
		orthonormalize();
	}
	
	/* Reconstruct orthonormal frame from X and Y axes.
	  Y is primary, X is secondary, Z is tertiary.
	*/
	void orthonormalize(void) {
		y=normalize(y);
		z=normalize(cross(x,y));
		x=normalize(cross(y,z));
	}
};

/* Orientation of our camera. Camera looks down +Z axis; Y points down. */
ortho_frame camera_frame;

/* Location of our camera */
vec3 camera_location=vec3(0,0,1.7);

/* Corresponding key is down */
bool key_is_down[256]={false};

/* Time it took to draw the last frame (for updating physics) */
float seconds_per_frame=0.01;

/* Nice handy timer: returns seconds */
double nice_timer(void) {return glutGet(GLUT_ELAPSED_TIME)*1.0e-3;}

/* Return a random float on [-1,+1] */
float rand_float(void) {
	return (rand()&0xffff)*(1.0/0xffff)*2.0-1.0;
}

/* Return a random two-vector on [-1,+1] */
vec3 rand_vec2(void) {
	return vec3(rand_float(),rand_float(),0.0);
}
/* Return the fractional part of this vector */
vec3 fract(const vec3 &v) {
	return v-vec3(floor(v.x),floor(v.y),floor(v.z));
}

void draw_sphere(const vec3 &pos,float r) {
	glPushMatrix();
	glTranslatef(pos.x,pos.y,pos.z);
	glutSolidSphere(r,12,8);
	glPopMatrix();
}

void my_display(void) {
	glFinish();
	int last_msec=glutGet(GLUT_ELAPSED_TIME);
	
	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
	glMatrixMode(GL_PROJECTION); /* set up projection matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
	static usgs_bil_terrain terrain("data/65458669/65458669");
	static bool first_time=true;
	if (first_time) {
		first_time=false;
		camera_location=terrain.get_xyz(terrain.w/2,terrain.h/2);
		camera_location.z+=100.0; /* float above terrain */
	}
	
	
/* Move the camera */
	float dt=seconds_per_frame;/* time between physics updates, seconds */
	
	float vel=walking_speed*350.0; /* velocity, in m/s */
	float forward_dir=0.0, strafe_dir=0.0;
	if (key_is_down['w']) {forward_dir+=1.0;}
	if (key_is_down['s']) {forward_dir-=1.0;}
	if (key_is_down['a']) {strafe_dir-=1.0;}
	if (key_is_down['d']) {strafe_dir+=1.0;}
	
	static vec3 camera_vel = vec3(0.0);
	float friction_frac=1.0;
	vec3 acc = (
		camera_frame.z*forward_dir+
		camera_frame.x*strafe_dir
		)*vel*dt - friction_frac*camera_vel;
	camera_vel = dt*acc + camera_vel;
	
	if (move_camera) 
		camera_location += camera_vel;

	glutPostRedisplay();
	
	camera_frame.orthonormalize();
	
/* Set up camera */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		glutGet(GLUT_WINDOW_WIDTH)/(float)glutGet(GLUT_WINDOW_HEIGHT), /* aspect ratio: width over height */
		0.1, /* Near clipping plane depth */
		100000.0 /* Far clipping plane depth */
	);
	gluLookAt(VEC3_TO_XYZ(camera_location),
		VEC3_TO_XYZ(camera_location + camera_frame.z), /* look down +Z */
		VEC3_TO_XYZ(-camera_frame.y) /* OpenGL +Y points up, our Y points down */
	);
	
	glMatrixMode(GL_MODELVIEW); /* set up model and view matrix */
	glLoadIdentity(); /* reset matrix at start of frame */
	
	glClearColor(0.3,0.5,0.7,0); /* set background color to blue */
	glClear(GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT);
	glEnable(GL_DEPTH_TEST);
	
	glColor4f(1,1,1,1);
	
	terrain.draw();
	
	

	/* Start the random tree sequence again */
	srand(10);
	for (int tree=0;tree<10000;tree++) {
		int x=rand()%terrain.w;
		int y=rand()%terrain.h;
		glColor3f(0.3,1.0,0.5);
		glPushMatrix();
		vec3 v=terrain.get_xyz(x,y);
		glTranslatef(VEC3_TO_XYZ(v));
		glutSolidCone(1.0, 10.0, 3,2);
		glPopMatrix();
	}
	
	glutSwapBuffers(); /* copy back buffer to front */
	
/* Statistics */
	glFinish();
	int cur_msec=glutGet(GLUT_ELAPSED_TIME); /* milliseconds since glutInit */
	int msec_elapsed=cur_msec-last_msec;
	printf("%d msec/frame; %.2f fps\n",
		msec_elapsed,1.0e3/msec_elapsed);
	last_msec=cur_msec;
	seconds_per_frame=msec_elapsed*1.0e-3;
}

int mouse_x, mouse_y;
void my_mouse(int state,int buttons,int x,int y) {
	mouse_x=x; mouse_y=y;  /* just stash motion */
}
void my_motion(int x,int y) { 
	float dx=x-mouse_x, dy=y-mouse_y; /* relative motion */
	
	float speed=1.0/100; /* radians rotation per pixel */
	camera_frame.nudge(speed*dx,speed*dy);
	
	mouse_x=x; mouse_y=y; 
	glutPostRedisplay();
}

void my_keyboard(unsigned char key,int mouse_x,int mouse_y) 
{
	//printf("Key '%c' going down\n",key);
	key_is_down[key]=true;
	glutPostRedisplay();
}
void my_keyboardup(unsigned char key,int mouse_x,int mouse_y) 
{
	//printf("Key '%c' going back up\n",key);
	key_is_down[key]=false;
}
int main(int argc,char *argv[]) {
	/* Start camera in reasonable orientation */
	camera_frame.y=vec3(0,0,-1); camera_frame.z=vec3(0,1,0);
	camera_frame.orthonormalize();

	glutInitWindowSize(800,600);
	glutInit(&argc,argv);
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
        glutInitDisplayMode (mode);
	int win=glutCreateWindow("Simple Interactive GUI Demo (try the WASD keys!)");
	glewInit();
	
	GLUI *glui = GLUI_Master.create_glui_subwindow(win, GLUI_SUBWINDOW_TOP);
	glui->set_main_gfx_window( win );
	
	new GLUI_Spinner(glui, "Speed (Mach): ", &walking_speed);
	
	glutDisplayFunc(my_display);
	glutMotionFunc(my_motion);
	glutMouseFunc(my_mouse);
	glutKeyboardFunc(my_keyboard);
	glutKeyboardUpFunc(my_keyboardup);
	
	glutMainLoop(); /* processes events */
	return 0;
}
